/*
 * Copyright (C) 1997-2005, R3vis Corporation.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA, or visit http://www.gnu.org/copyleft/lgpl.html.
 *
 * Original Contributor:
 *   Wes Bethel, R3vis Corporation, Marin County, California
 * Additional Contributor(s):
 *
 * The OpenRM project is located at http://openrm.sourceforge.net/.
 */
/*
 * $Id: rmarray.c,v 1.1 2005/02/19 16:36:40 wes Exp $
 * Version: $Name: OpenRM-1-6-0-RC5 $
 * $Revision: 1.1 $
 * $Log: rmarray.c,v $
 * Revision 1.1  2005/02/19 16:36:40  wes
 * Distro merge and consolidation.
 *
 */

#include <rm/rm.h>
#include "rmprivat.h"

static RMenum
private_rmArrayRealloc(RMarray *a)
{
    char *t;
    size_t newSize;

    newSize = a->currentArraySize*a->elementSize + a->elementSize * a->chunkSize;

#if 1
    t = (void *)realloc(a->data, newSize);
    if (t == NULL)
    {
	char buf[512];
	sprintf(buf, "private_rmArrayRealloc error: unable to realloc an array of size %ld bytes. ", (long)newSize);
	rmError(buf);
	return RM_WHACKED;
    }
#else
    /* manual realloc */
    t = (char *)malloc(newSize);
    memcpy(t, a->data, a->currentArraySize*a->elementSize);
    free((void *)(a->data));
#endif
    a->data = t;
    a->currentArraySize += a->chunkSize;
    
    return RM_CHILL;
}

RMenum
rmArrayDelete(RMarray **a)
{
    void *data;
    
    if ((RM_ASSERT(a, "rmArrayDelete error - the input array object handle is NULL.") == RM_WHACKED) ||
	(RM_ASSERT(*a,"rmArrayDelete error - the input array object is NULL") == RM_WHACKED))
	return RM_WHACKED;

    data = (*a)->data;

    if (data != NULL)
	free(data);

    free((void *)*a);
    *a = NULL;

    return RM_CHILL;
}

RMarray *
rmArrayNew(int initSize,
	   int chunkSize,
	   int elementSize)
{
    /* creates a new array object  */
/*    RMarray *t = (RMarray *)calloc(1, sizeof(RMarray)); */
    RMarray *t = (RMarray *)malloc(sizeof(RMarray));
    memset(t, 0, sizeof(RMarray));
    
    t->nItems = 0;
    t->chunkSize = chunkSize;
    t->elementSize = elementSize;
    t->currentArraySize = initSize;

    if (initSize != 0)
    {
	t->data = (char *)malloc(sizeof(char)*(initSize* elementSize));
	memset(t->data, 0, sizeof(char)*(initSize* elementSize));
/*	t->data = (char *)calloc(initSize, elementSize); */
    }

    return t;
}

void *
rmArrayGet(const RMarray *toQuery,
	   int indx)
{
    /* get a handle to an item in the array */
    off_t offset;

    char *t;

    if (RM_ASSERT(toQuery, "rmArrayGet error - the input RMarray object is NULL.") == RM_WHACKED)
	return NULL;

    if (indx >= toQuery->nItems)
    {
	char buf[512];
	sprintf(buf,"rmArrayGet warning - request for item #%d, but there are only %d items in the array.", indx, toQuery->nItems);
	return NULL;
    }

    t = toQuery->data;
    offset = indx * toQuery->elementSize;
    return (void *)(t+offset);
}

RMenum
rmArrayAdd(RMarray *toModify,
	   void *newData)
{
    off_t offset;
    char *dest;
    
    if (RM_ASSERT(toModify, "rmArrayAdd error - the input RMarray object is NULL.") == RM_WHACKED)
	return RM_WHACKED;
    
    /* see if there's enough room to add one more thingamajig to the array */
    if ((toModify->nItems + 1) >= toModify->currentArraySize)
    {
	if (private_rmArrayRealloc(toModify) != RM_CHILL)
	{
	    rmError("rmArrayAdd() - unable to realloc sufficient space to add more items to the input array. Action fails. ");
	    return RM_WHACKED;
	}
    }
    
    offset = toModify->nItems * toModify->elementSize;
    dest = (char *)(toModify->data) + offset;
    memcpy((void *)dest, newData, toModify->elementSize);

    toModify->nItems += 1;

    return RM_CHILL;
}

int
rmArrayNumItems(const RMarray *toQuery)
{
    if (RM_ASSERT(toQuery, "rmArrayNumItems error - the input array object is NULL") == RM_WHACKED)
	return -1;
    
    return toQuery->nItems;
}

#if 0
/* we might implement this if needed */
RMenum
rmArraySet(RMarray *toModify,
	   int indx,
	   void *data)
{
    /* set an item in the array */
}
#endif

void
rmArraySort(RMarray *a,
	    int (*compareFunc)(const void *a, const void *b))
{
    int nItems = rmArrayNumItems(a);
    qsort(a->data, nItems, a->elementSize, compareFunc);
}

/* EOF */
